/* 

  HEV General Utilities

  These source files are used by other programs in the NIST MSCD SAVG
  High-end Visulaization software suite- they are generally symlinked from
  the particular projects's directory to the $HEVROOT/src/genUtils
  directory.  When creating tarballs for release, symlinks are dereferenced
  so a complete package is contained in the tarball.

  Authors:
      John Kelso, Adele Peskin, Steve Satterfield
      Scientific Applications & Visualization Group
           SAVG web page: http://math.nist.gov/mcsd/savg/
      Mathematical & Computational Science Division
           MCSD web page: http://math.nist.gov/mcsd/
      National Institute of Standards and Technology
           NIST web page: http://www.nist.gov/


  For information about this software contact John Kelso, kelso@nist.gov

  "This software was developed at the National Institute of Standards and
  Technology by employees of the Federal Government in the course of their
  official duties. Pursuant to title 17 Section 105 of the United States
  Code this software is not subject to copyright protection and is in the
  public domain. This is an experimental system.  NIST assumes no
  responsibility whatsoever for its use by other parties, and makes no
  guarantees, expressed or implied, about its quality, reliability, or any
  other characteristic.

 We would appreciate acknowledgement if the software is used."

*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/*
  pass a line and break it into argc and **argc sort of like a shell does.
  stuff within "" or '' is treated as one parameter. the \ character will
  escape the next character.  A # throws away the rest of the line.

  DOES NOT DO ERROR CHECKING FOR TERMINATING QUOTES- AN EOL OR COMMENT
  TERMINATES THE QUOTE.
*/

static void dpf_charsToArgcArgv_addChar(char c, int *pargc, char ***pargv, int *inSpace) {
  if (*inSpace) {
    // increment number of parameters
    (*pargc)++ ;
    
    // add one more array element
    *pargv = (char**) realloc(*pargv, (*pargc)*(sizeof(char *))) ;
    
    // create a null string
    char *hold = (char *) malloc(sizeof(char)) ;
    hold[0] = '\0' ;
    (*pargv)[(*pargc)-1] = hold ;
    *inSpace = 0 ;
  }

  int i = (*pargc)-1 ; 

  // one for new character, one for terminator
  int len = strlen((*pargv)[i])+2 ;
  char *hold = (char *) malloc(len*sizeof(char)) ;
  strcpy(hold, (*pargv)[i]) ;
  hold[len-2] = c ;
  hold[len-1] = '\0' ;
  free((*pargv)[i]) ;
  (*pargv)[i] = hold ;

}

void dpf_charsToArgcArgv(char *line, int *pargc, char ***pargv) {
  // reset each line
  // free up memory, set argc to zero, set argv to NULL
  *pargc = 0 ;
  *pargv = NULL ;

  int inSpace = 1 ;
  int inDoubleQuote = 0 ;
  int inSingleQuote = 0 ;
  int prevEscape = 0 ;

  for (unsigned int i=0; i<strlen(line); i++) {
    char c = line[i] ;
    if (c == '\n')
      continue;
    if  (c == '\\') {
      prevEscape = 1 ;
    } else {
      if (prevEscape) {
	// if previous character was an escape, just add to arg
	dpf_charsToArgcArgv_addChar(c, pargc, pargv, &inSpace) ;
      } else if (c == '"') {
	if (inSingleQuote) {
	  dpf_charsToArgcArgv_addChar(c, pargc, pargv, &inSpace) ;
	} else 
	  if (inDoubleQuote) {
	    inDoubleQuote = 0 ;
	  } else {
	    inDoubleQuote = 1 ;
	  }
      } else if  (c == '\'') {
	if (inDoubleQuote)
	  dpf_charsToArgcArgv_addChar(c, pargc, pargv, &inSpace) ;
	else 
	  if (inSingleQuote) {
	    inSingleQuote = 0 ;
	  } else {
	    inSingleQuote = 1 ;
	  }

      } else if  (c == ' ') {
	if (prevEscape || inSingleQuote || inDoubleQuote) {
	  dpf_charsToArgcArgv_addChar(c, pargc, pargv, &inSpace) ;
	} else if (!inSpace) {
	  inSpace = 1 ;
	}

      } else if (c == '#') {
	return ;

      } else {
	dpf_charsToArgcArgv_addChar(c, pargc, pargv, &inSpace) ;
      }
      prevEscape = 0 ;
    }
  }
}
